extern crate futures;
extern crate libc;
extern crate tokio;

use futures::{Future, Stream};
use std::io::{self, Read, Write};
use std::net::Shutdown;
use std::os::unix::io::AsRawFd;
use tokio::io::{AsyncRead, AsyncWrite};
use tokio::net::{TcpListener, TcpStream};

pub trait HasShutdown {
    fn shutdown(&mut self, how: Shutdown) -> io::Result<()>;
}

impl HasShutdown for TcpStream {
    fn shutdown(&mut self, how: Shutdown) -> io::Result<()> {
        tokio::net::TcpStream::shutdown(&self, how)
    }
}

impl<S: AsRawFd> AsRawFd for ClosingStream<S> {
    fn as_raw_fd(&self) -> i32 {
        self.0.as_raw_fd()
    }
}

pub struct ClosingStream<S>(pub S);

impl<S> Read for ClosingStream<S>
where
    S: Read,
{
    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        self.0.read(buf)
    }
}

impl<S> Write for ClosingStream<S>
where
    S: Write,
{
    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
        self.0.write(buf)
    }

    fn flush(&mut self) -> io::Result<()> {
        self.0.flush()
    }
}

impl<S> AsyncRead for ClosingStream<S> where S: AsyncRead {}

impl<S> AsyncWrite for ClosingStream<S>
where
    S: Write + HasShutdown,
{
    fn shutdown(&mut self) -> futures::Poll<(), io::Error> {
        println!("shutdown called on closing stream");
        HasShutdown::shutdown(&mut self.0, Shutdown::Write)?;
        Ok(futures::Async::Ready(()))
    }
}

fn do_work(
    io: impl AsyncRead + AsyncWrite + AsRawFd,
    side: String,
    file: String,
) -> impl Future<Item = (), Error = ()> {
    use std::ffi::CString;

    let fd = io.as_raw_fd();
    let (rh, wh) = io.split();
    let side1 = side.clone();
    let side2 = side.clone();
    let side3 = side.clone();
    let now1 = std::time::Instant::now();
    let now2 = now1;

    tokio::fs::File::open(file)
        .map_err(|e| panic!("open null error {:?}", e))
        .and_then(move |f| {
            tokio::io::copy(f, wh).and_then(move |(_, _, wh)| {
                println!("finish file copy: {:?}", std::time::Instant::now() - now2);
                tokio::io::shutdown(wh)
            })
        })
        .map_err(move |e| {
            eprintln!("{} copy err: {:?}", side3, e);
            e
        })
        .join(
            tokio::fs::OpenOptions::new()
                .write(true)
                .open("/dev/null")
                .map_err(|e| panic!("open error {:?}", e))
                .and_then(move |f| {
                    tokio::io::copy(rh, f)
                        .map(move |_| {
                            println!("finish socket copy: {:?}", std::time::Instant::now() - now1);
                            unsafe {
                                let ret = libc::shutdown(fd, libc::SHUT_RD);
                                if ret != 0 {
                                    let e = CString::new("shutdown read")
                                        .expect("creating error string");
                                    libc::perror(e.as_ptr());
                                }
                            }
                        })
                        .map_err(|e| {
                            println!("copy error: {:?}", e);
                            e
                        })
                })
                .map_err(move |e| {
                    eprintln!("{} writing to null err: {:?}", side, e);
                    e
                }),
        )
        .map(move |_| eprintln!("finish {}", side1))
        .map_err(move |e| eprintln!("finish {} err: {:?}", side2, e))
}

fn main() {
    let input = std::env::args().nth(1).unwrap_or("file-100m".to_owned());
    let server_input = std::env::args().nth(2).unwrap_or("file-100m".to_owned());
    let input1 = input.clone();

    let listener =
        TcpListener::bind(&"127.0.0.1:0".parse::<std::net::SocketAddr>().unwrap()).unwrap();
    let local_addr = listener.local_addr().unwrap();

    let mut runtime = tokio::runtime::Runtime::new().unwrap();

    runtime.spawn(
        listener
            .incoming()
            .map_err(|_| eprintln!("error accepting sockets"))
            .for_each(move |s| {
                tokio::spawn(do_work(
                    ClosingStream(s),
                    "server".to_owned(),
                    server_input.clone(),
                ));
                Ok(())
            }),
    );

    std::thread::sleep(std::time::Duration::from_millis(100));

    for n in 1..5 {
        let input1 = input1.clone();
        let client = format!("client-{}", n);
        runtime.spawn(
            tokio::net::TcpStream::connect(&local_addr)
                .map_err(|e| eprintln!("connect error: {}", e))
                .and_then(move |c| do_work(ClosingStream(c), client, input1))
                .map(|_| ()),
        );
    }

    runtime.shutdown_on_idle().wait().unwrap();
}
